/* Module   : ScrollContainer.java
** Abstract : Implementation of scrollable container for CHUI
**
** Copyright (c) 2005-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description------------------------------
** 001 EVL 20050405   @20656 Creation based on CHARVA JPanel class.
**                           Necessary modification to get compiled
** 002 EVL 20050408   @20671 Adding contents scrolling generation
** 003 EVL 20050421   @20831 Changing constructor orders
** 004 EVL 20050424   @20860 Adding scrollable flag
** 005 EVL 20050520   @21252 Adding function to get the visible size
** 006 EVL 20050531   @21337 Changing the inner variable names to sync
**                           with the rest of the project and preparing
**                           to the javadoc processing
** 007 EVL 20050607   @21435 Syntax for comments checking
** 008 EVL 20050729   @21915 Removing the ugly tabulation from source.
** 009 SIY 20050831   @22547 Fixed formatting and spelling.
** 010 NVS 20060227   @24870 Key actions KA_* are used instead of virtual
**                           keys VK_* for all editing operations.
** 011 SIY 20061220   @31723 Fixed handling of the scrolling of scrollable
**                           containers.
** 012 SIY 20081125   @40692 Fixed scrolling handling.
** 013 SIY 20081125   @40701 Fixed regression caused by #012. Scrolling to
**                           topmost widget, rather than bottommost is 
**                           necessary for proper initial appearance of 
**                           frame.
** 014 SIY 20090409   @41737 If widget which should be made visible is a
**                           label, ignore its length.
** 015 SIY 20100619          Switching to own implementation of character UI.
** 016 SIY 20100804          Changes in imports, some refactorings.
** 017 SIY 20100915          locateFrame() moved to UiUtils.
** 018 SIY 20101122          Switching from KeyEvent to KeyInput.
** 019 SIY 20111202          Refactored and moved from CHUI package.
** 020 CA  20140805          Added generics to force the abstract widget classes to not use
**                           implementation-specific code: the abstract widgets will have access
**                           only to the base APIs provided by the OutputManager.
** 021 HC  20140911          GUI coordinate support - data types refactoring.
** 022 CA  20140926          Added shared widget configuration support, to allow GUI/ChUI
**                           concrete implementation for the same widget ID.  Refactored the
**                           widget configuration classes: all fields were made public; these
**                           classes need to be as dumb as possible, all logic related to
**                           setting/getting a certain field should be at the widget or at the
**                           caller. Refs #2254
** 023 HC  20141021          Implemented window move/position support.
** 024 IAS 20150219          Changed ensureVisibility() method implementation.
*/

package com.goldencode.p2j.ui.client;

import java.util.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.event.listener.*;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * A ScrollContainer is a scrollable container for Frame, ChildFrame and
 * DialogBox classes.
 * TODO<SIY>: this class does not contain draw() method, so there is no 
 * UI-specific versions, but {@link #ensureVisibility(Widget)} method should be
 * carefully reviewed, perhaps it requires some adjustments in order to support 
 * GUI properly.
 */
public class ScrollContainer<O extends OutputManager<?>>
extends BorderedPanel<O>
implements Scrollable
{
   // Instance members
   /** ScrollListeners vector. */
   private ArrayList<ScrollListener> scrollListeners = null;

   /** The preferred row size for the container. */
   private double preferredRows = 25;
   
   /** The preferred column size for the container. */
   private double preferredColumns = 40;
   
   /** The origin of this container. */
   private Point origin = new Point(0,0);
   
   /** Track first time when widget is displayed. */
   private boolean first_view = true; 

   /** Source scroll container used to retrieve some useful info. */
   private Frame<O> frame = null;

   // Class methods
   /**
    * The default constructor sets the layout manager to NullLayout.
    * 
    * @param   isScrollable
    *          The boolean flag that indicates whether the ScrollContainer has
    *          the scrolling capability.
    */
   public ScrollContainer(boolean isScrollable, Frame<O> frame)
   {
      frame.config().scrollable = isScrollable;
      this.frame = frame;
   }

   // New/overridden methods
   /**
    * Setting the visible size for the container.
    * 
    * @param   width
    *          The initial value for the visible width of the container.
    * @param   height
    *          The initial value for the visible height of the container.
    */
   public void setVisibleSize(double width, double height)
   {
      width = Coordinate.scale(width);
      height = Coordinate.scale(height);
      
      preferredRows = height;
      preferredColumns = width;

      Dimension d = dimension();

      // safe, comparison of scaled doubles
      if (width > d.width)
         d.width = width;
      if (height > d.height)
         d.height = height;
      
      setSize(d);
   }

   /**
    * @see com.goldencode.p2j.ui.client.widget.Widget#setVisible(boolean)
    */
   @Override
   public void setVisible(boolean visible)
   {
      if (first_view)
      {
         Widget<O>[] list = widgets(); 
         
         Widget<O> target = null;
         
         for (Widget<O> comp : list)
         {
            if (target == null)
            {
               target = comp;
               continue;
            }
            
            double y1 = Coordinate.scale(target.location().y + target.height());
            double y2 = Coordinate.scale(comp.location().y + comp.height());

            if (y1 > y2)
            {
               target = comp;
            }
         }
         
         if (target != null)
         {
            ensureVisibility(target);
         }
         
         first_view = false;
      }
      
      super.setVisible(visible);
   }
   
   /**
    * Getting the visible size of the ScrollContainer.
    * 
    * @return  The Dimension object which represents the size of the container.
    */
   public Dimension getVisibleSize()
   {
      return new Dimension(preferredColumns, preferredRows);
   }
   
   /**
    * Handling the key event. It is required to intercept this handler to
    * adjust the ScrollContainer contents after moving focus to the next
    * widget which can currently be out of the visible part of the container.
    * In this case the contents is being scrolled to be sure the currently
    * focused component is always visible.
    * 
    * @param   keyEvent
    *          The KeyEvent object to be handled within the ScrollContainer.
    */
   @Override
   public void processKeyEvent(KeyInput keyEvent)
   {
      /*
       * First call all KeyListener objects that may have been registered for
       * this component.
       */
      super.processKeyEvent(keyEvent);

      /*
       * Check if any of the KeyListeners consumed the KeyEvent.
       */
      if (keyEvent.isConsumed())
         return;

      int key = keyEvent.actionCode();

      if (!frame.config().scrollable || !isScrollable())
         return;

      if (key == Keyboard.KA_TAB          || 
          key == Keyboard.KA_BACK_TAB     || 
          key == Keyboard.KA_CURSOR_DOWN  || 
          key == Keyboard.KA_CURSOR_UP    || 
          key == Keyboard.KA_CURSOR_LEFT  || 
          key == Keyboard.KA_CURSOR_RIGHT || 
          key == Keyboard.KA_LEFT_END     || 
          key == Keyboard.KA_RIGHT_END)
      {
         if (ensureVisibility(currentFocus()))
            repaint();
      }
   }

   /**
    * Make sure that given component is visible in the viewport.
    * 
    * @param   comp
    *          The component which should be made visible.
    *          
    * @return  <code>true</code> if there was scrolling.
    * 
    */
   boolean ensureVisibility(Widget<O> comp)
   {
      if (!isScrollable())
      {
         return false;
      }

      Point     p = comp.location();
      Dimension d = comp.dimension();

      if (comp instanceof Label)
         d.width = 1;

      Rectangle r = comp.getVisiblePart();
      if (r != null)
      {
         // use new logic only for widgets implementing
         // sensible getVisiblePart() method to retain 
         // backward compatibility
         return ensureVisibility(p, d, r);
      }
      
      boolean scrolled = false;
      
      // DOWN_RIGHT
      if (p.x - origin.x < 0 && p.y - origin.y < 0)
      {
         origin.x = p.x;
         origin.y = p.y;
         EventManager.postEvent(new ScrollEvent(this, 
                                                ScrollDirection.DOWN_RIGHT,
                                                new Point(p.x, p.y)));
         scrolled = true;
      }
      // DOWN
      else if (p.x - origin.x + d.width - 1 >= 0               && 
               p.x - origin.x + d.width - 1 < preferredColumns && 
               p.y - origin.y < 0)
      {
         origin.y = p.y;
         EventManager.postEvent(new ScrollEvent(this, 
                                                ScrollDirection.DOWN,
                                                new Point(0, p.y)));
         scrolled = true;
      }
      // DOWN_LEFT
      else if (p.x - origin.x + d.width - 1 >= preferredColumns &&
               p.y - origin.y < 0)
      {
         origin.x = p.x + d.width - 1 - preferredColumns;
         origin.y = p.y;
         EventManager.postEvent(new ScrollEvent(this, 
                                                ScrollDirection.DOWN_LEFT,
                                                new Point(p.x + d.width - 1, p.y)));
         scrolled = true;
      }
      // RIGHT
      else if (p.x - origin.x < 0 && p.y - origin.y >= 0 &&
               p.y - origin.y < preferredRows)
      {
         origin.x = p.x;
         EventManager.postEvent(new ScrollEvent(this, 
                                                ScrollDirection.RIGHT,
                                                new Point(p.x, 0)));
         scrolled = true;
      }
      // LEFT
      else if (p.x - origin.x + d.width - 1 >= preferredColumns &&
               p.y - origin.y >= 0                              && 
               p.y - origin.y < preferredRows)
      {
         origin.x = p.x + d.width - 1 - preferredColumns;
         EventManager.postEvent(new ScrollEvent(this, 
                                                ScrollDirection.LEFT,
                                                new Point(p.x + d.width - 1, 0)));
         scrolled = true;
      }
      // UP_RIGHT
      else if (p.x - origin.x < 0 &&
               p.y - origin.y + d.height - 1 >= preferredRows)
      {
         origin.x = p.x;
         origin.y = p.y + d.height - 1 - preferredRows;
         EventManager.postEvent(new ScrollEvent(this, 
                                                ScrollDirection.UP_RIGHT,
                                                new Point(p.x, p.y + d.height - 1)));
         scrolled = true;
      }
      // UP
      else if (p.x - origin.x + d.width - 1 >= 0               &&
               p.x - origin.x + d.width - 1 < preferredColumns &&
               p.y - origin.y + d.height - 1 >= preferredRows)
      {
         origin.y = p.y + d.height - 1 - preferredRows;
         EventManager.postEvent(new ScrollEvent(this, 
                                                ScrollDirection.UP,
                                                new Point(0, p.y + d.height - 1)));
         scrolled = true;
      }
      // UP_LEFT
      else if (p.x - origin.x + d.width - 1 >= preferredColumns &&
               p.y - origin.y + d.height - 1 >= preferredRows)
      {
         origin.x = p.x + d.width - 1 - preferredColumns;
         origin.y = p.y + d.height - 1 - preferredRows;
         EventManager.postEvent(new ScrollEvent(this, 
                                                ScrollDirection.UP_LEFT,
                                                new Point(p.x + d.width - 1,
                                                          p.y + d.height - 1)));
         scrolled = true;
      }
      
      return scrolled;
   }

   // Scrollable interface methods
   /**
    * Register a ScrollListener object for this ScrollContainer.
    * 
    * @param   scrollListener
    *          The new scroll listener to be registered for this container.
    */
   @Override
   public void addScrollListener(ScrollListener scrollListener)
   {
      if (scrollListeners == null)
         scrollListeners = new ArrayList<ScrollListener>();
      
      scrollListeners.add(scrollListener);
   }

   /**
    * Process scroll events generated by this ChuiContainer.
    * 
    * @param   scrollEvent
    *          The scroll event to be processed.
    */
   @Override
   public void processScrollEvent(ScrollEvent scrollEvent)
   {
      if (scrollListeners != null)
      {
         for (ScrollListener sl : scrollListeners)
         {
            sl.scroll(scrollEvent);
         }
      }
   }

   /** 
    * Returns the preferred size of the viewport for this Container when it is
    * in a ScrollPane (this method implements the Scrollable interface). The
    * size is determined by the number of rows and columns set for this
    * DialogBox (either in the constructor or in the setColumns() and
    * setRows() methods).
    * 
    * @return  The Dimension object which represents the preferred size of
    *          the container.
    */
   @Override
   public Dimension preferredSize()
   {
      return new Dimension(preferredColumns, preferredRows);
   }

   /**
    * @see com.goldencode.p2j.ui.client.widget.Container#setFocus(Widget)
    */
   @Override
   public void setFocus(Widget<O> focus)
   {
      boolean needRepaint = false;
      
      if (belongsTo(focus))
         needRepaint = ensureVisibility(focus);
      
      super.setFocus(focus);
      
      if (needRepaint)
      {
         @SuppressWarnings("unchecked")
         Widget<O> frame = (Widget<O>) UiUtils.locateFrame(this);
         
         if (frame != null)
            frame.repaint();
      }
   }
   
   /**
    * Get the frame associated with this scroll container.
    * 
    * @return   See above.
    */
   public Frame<O> getFrame()
   {
      return frame;
   }
   
   /**
    * Make sure that given component is visible in the viewport.
    * 
    * @param   p
    *          The component's location.
    * @param   d
    *          The component's dimension.
    * @param   r
    *          The component's sub-rectangle which should be visible as much as possible.
    *          
    * @return  <code>true</code> if there was scrolling.
    * 
    */
   protected boolean ensureVisibility(Point p, Dimension d, Rectangle r)
   {
      double x1 = p.x + r.left();
      double y1 = p.y + r.top();
      double x2 = p.x + r.right();
      double y2 = p.y = r.bottom();
      
      double dx = shift4overlap(x1, x2, origin.x, origin.x + preferredColumns);
      double dy = shift4overlap(y1, y2, origin.y, origin.y + preferredRows);

      ScrollDirection direction = ScrollDirection.of(-dx, -dy);
      if (direction == null )
      {
         return false;
      }

      origin.x -= dx;
      origin.y -= dy;

      Point norigin = new Point(origin);
      if (ScrollDirection.up(direction))
      {
         norigin.y += preferredRows - 1;
      }
      if (ScrollDirection.left(direction))
      {
         norigin.x += preferredColumns - 1;
      }
      EventManager.postEvent(new ScrollEvent(this, 
            direction,
            new Point(norigin)));
      return true;
   }

   /**
    * Find min shift offset to achieve max overlap of two line segments  
    *
    * @param x1
    *       start of the first segment
    * @param x2
    *       end of the first segment, x1 < x2
    * @param t1
    *       start of the second segment
    * @param t2
    *       end of the second segment, t1 < 21
    * @return
    *    offset
    */
   private double shift4overlap(double x1, double x2, double t1, double t2)
   {
      return x2 - x1 < t2 - t1? shiftInside(x1, x2, t1, t2) : -shiftInside(t1, t2, x1, x2); 
   }

   /**
    * Find min shift offset to out first line segment inside the second one  
    *
    * @param x1
    *       start of the first segment
    * @param x2
    *       end of the first segment, x1 < x2
    * @param t1
    *       start of the second segment
    * @param t2
    *       end of the second segment, t1 < x2,
    *       
    * @return
    *    offset x2 - x1 <= t2 - t1
    */
   private double shiftInside(double x1, double x2, double t1, double t2)
   {
      /*
       * t1 <= x1 + dx && x2 + dx <= t2  
       */
      return x1 < t1 ? t1 - x1 : t2 < x2 ? t2 - x2 : 0; 
   }

   /**
    * Check if container can be scrolled.
    * 
    * @return  <code>true</code> if container can be scrolled.
    */
   private boolean isScrollable()
   {
      Dimension actual  = dimension();
      Dimension visible = getVisibleSize();
      
      if (actual.width > visible.width || actual.height > visible.height)
         return true;
      
      return false;
   }

}
